#pragma once

#include "lab.h"

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <float.h>
#include <assert.h>
#include <math.h>

#define crossproduct(v1, v2, res) \
  (res)[0] = (v1)[1]*(v2)[2] - (v2)[1]*(v1)[2];\
  (res)[1] = (v1)[2]*(v2)[0] - (v2)[2]*(v1)[0];\
  (res)[2] = (v1)[0]*(v2)[1] - (v2)[0]*(v1)[1]

#define dotproduct(u, v) ((u)[0]*(v)[0] + (u)[1]*(v)[1] + (u)[2]*(v)[2])

static inline void normalise(float *f)
{
  const float len = 1.0f/sqrtf(dotproduct(f, f));
  for(int k=0;k<3;k++) f[k] *= len;
}


typedef struct gamutmap_t
{
  int64_t wd, ht;
  float *max;
}
gamutmap_t;

static inline int gamutmap_init(
    gamutmap_t *data,
    const char *filename,
    int channel)
{
  // read greyscale pfm of maximum per x/y chromaticity coordinate
  data->wd = data->ht = 0;
  data->max = 0;
  FILE *f = fopen(filename, "rb");
  if(f)
  {
    uint64_t w, h;
    if(fscanf(f, "PF\n%ld %ld\n%*[^\n]", &w, &h) != 2) goto error;
    data->wd = w;
    data->ht = h;
    (void)fgetc(f);
    data->max = malloc(sizeof(float)*(data->wd*data->ht));
    for(int64_t k=0;k<data->wd*data->ht;k++)
    {
      float col[3];
      fread(col, sizeof(float), 3, f);
      data->max[k] = col[channel];
    }
    fclose(f);
    return 0;
  }
error:
  if(f) fclose(f);
  return 1;
}

static inline void gamutmap_cleanup(
    gamutmap_t *data)
{
  data->wd = data->ht = 0;
  free(data->max);
}

// returns the maximum value of a smooth spectrum for which the
// given xyz coordinate is a metamer. for valid reflectances,
// this returns <= 1.0.
static inline float gamutmap_get_max(
    const gamutmap_t *data,
    const float *xyz)
{
  const float brightness = xyz[0] + xyz[1] + xyz[2];
  float x = xyz[0] / brightness;
  float y = xyz[1] / brightness;

  // could interpolate this linearly i guess:
  int i = fmaxf(0.0f, fminf(data->wd-1, floorf(x * data->wd)));
  int j = fmaxf(0.0f, fminf(data->ht-1, floorf(y * data->ht)));
  assert(data->max);
  return brightness / data->max[data->wd * j + i];
}

static inline float gamutmap_get_max_bilin(
    const gamutmap_t *data,
    const float x,
    const float y)
{
  const int i = (int)x;
  const int j = (int)y;
  const float fi = x-i;
  const float fj = y-j;
  const float v00 = data->max[data->wd * (j  ) + i];
  const float v01 = data->max[data->wd * (j  ) + i+1];
  const float v10 = data->max[data->wd * (j+1) + i];
  const float v11 = data->max[data->wd * (j+1) + i+1];
  const float v = v00*(1.f-fi)*(1.f-fj) + v01*fi*(1.f-fj) + v10*(1.f-fi)*fj + v11*fi*fj;
  return 1.f / v;
}

static inline int gamutmap_get_normal(
    const gamutmap_t *data,
    const float xyz_in[3],
    float normal[3])
{
  // construct some tangent frame in Lab around the
  // given point in xyz. to do that, we convert the point to xy and
  // query the max in the neighbouring pixels, convert to Lab and compute
  // the tangents there.
  const float di = 1.0f/data->wd, dj = 1.0f/data->ht;
  const float brightness = xyz_in[0] + xyz_in[1] + xyz_in[2];
  float x = xyz_in[0] / brightness;
  float y = xyz_in[1] / brightness;
  const float delta = 2.0f;
  float i = fmaxf(delta, fminf(data->wd-2*delta, x * data->wd));
  float j = fmaxf(delta, fminf(data->ht-2*delta, y * data->ht));
  float off[4][2] = {{-delta,0},{delta,0},{0,-delta},{0,delta}};
  float Lab[4][3];
  for(int k=0;k<4;k++)
  {
    // convert xyY to XYZ, then scale by max brightness
    const float xyz[3] = {x + off[k][0]*di, y + off[k][1]*dj, 1.0f - (x+off[k][0]*di) - (y+off[k][1]*dj)};
    const float m = gamutmap_get_max_bilin(data, i+off[k][0], j+off[k][1]);// 106.8f * data->max[data->wd * (j+off[k][1]) + i+off[k][0]];
    if(!(m > 1e-5f)) return 1;
    // const float XYZ[3] = { xyz[0] * brightness / m, xyz[1] * brightness / m, xyz[2] * brightness / m };
    const float XYZ[3] = { xyz[0] / m, xyz[1] / m, xyz[2] / m };
    assert(xyz[0] == xyz[0]);
    assert(xyz[1] == xyz[1]);
    assert(xyz[2] == xyz[2]);
    dt_XYZ_to_Lab(XYZ, Lab[k]);
    // fprintf(stdout, "%d %g %g %g %g %g %g\n", 0, Lab[k][0], Lab[k][1], Lab[k][2], 0.0, 0.0, 0.0);
    assert(Lab[k][0] == Lab[k][0]);
    assert(Lab[k][1] == Lab[k][1]);
    assert(Lab[k][2] == Lab[k][2]);
  }
  // compute tangent vectors:
  float dx[3], dy[3];
  for(int k=0;k<3;k++) dx[k] = Lab[1][k] - Lab[0][k];
  for(int k=0;k<3;k++) dy[k] = Lab[3][k] - Lab[2][k];
  normalise(dx);
  if(!(dx[0] == dx[0])) return 1;
  normalise(dy);
  if(!(dy[0] == dy[0])) return 1;
  crossproduct(dx, dy, normal);
  normalise(normal);
  if(!(normal[0] == normal[0])) return 1;
  assert(normal[0] == normal[0]);
  assert(normal[1] == normal[1]);
  assert(normal[2] == normal[2]);
  return 0;
}

// perform gamut mapping XYZ -> XYZ within the gamut of valid reflectances.
static inline int gamutmap(
    const gamutmap_t *data,
    const float *in_xyz,
    float *out_xyz)
{
  float in_lab[3], lab[3], xyz[3], max;
  // need to iterate a little here.
  // want to be perpendicular to lab gamut, so do a gradient descent/newton here
  max = gamutmap_get_max(data, in_xyz);
  // outside spectral locus
  if(max <= 0.0f)
  {
    for(int k=0;k<3;k++) out_xyz[k] = 0.0f;
    return 0;
  }
  if(max <= 1.0f)
  { // all good, valid reflectance
    for(int k=0;k<3;k++) out_xyz[k] = in_xyz[k];
    return 0;
  }

#ifdef DEBUG_SCALE
  // XXX DEBUG: only perform direct scaling and no iteration
  for(int k=0;k<3;k++) out_xyz[k] = in_xyz[k] / max;
  return 0;
#endif

  dt_XYZ_to_Lab(in_xyz, in_lab);
  for(int k=0;k<3;k++) xyz[k] = in_xyz[k];

  // project to bring us to the boundary of the reflectance gamut,
  // same chromaticity (but at first large delta E)
  for(int k=0;k<3;k++) xyz[k] = xyz[k] / max;
  dt_XYZ_to_Lab(xyz, lab);
  float d[3], n[3];
  for(int k=0;k<3;k++) d[k] = lab[k] - in_lab[k];
  float dot_dd = dotproduct(d, d);
  if(gamutmap_get_normal(data, xyz, n))
  {
      // fail
      fprintf(stderr, "diverged %g, %g, %g\n", in_xyz[0], in_xyz[1], in_xyz[2]);
      for(int k=0;k<3;k++) out_xyz[k] = 0.0f;
      return 1;
  }
  float dot_dn = dotproduct(d, n);

  float tmp[3];
  int it = 0;
  float step = 1.;//.5f;
  float last_deltaE2 = FLT_MAX;
  while(1)
  {
    if(it++ > 100) {
        // fail
      fprintf(stderr, "diverged %g, %g, %g\n", in_xyz[0], in_xyz[1], in_xyz[2]);
      for(int k=0;k<3;k++) out_xyz[k] = 0.0f;
      return 1;
    }
    if(step < 1e-2) break; // that counts as success :(

    // check whether we converged:
    const float cos_err = fabsf(dot_dn)/sqrtf(dot_dd);
    assert(cos_err == cos_err);

    // converged?
    if(fabsf(cos_err - 1.0f) < 1e-2f) break;
    else if(dot_dd > last_deltaE2)
    { // error increased, decrease stepsize
      step *= .5f;
    }
    // else step = fminf(1.0f, step * 2.0f);

    // predict where on the gamut boundary we would have a shorter distance d:
    const float t = dot_dn;// /dot_nn;
    int ii = 0;
    while(1)
    {
      float newlab[3];
      for(int k=0;k<3;k++) newlab[k] = lab[k] * (1.0f-step) + (in_lab[k] + t * n[k]) * step;
      // go back to xyz for next iteration
      dt_Lab_to_XYZ(newlab, tmp);
      max = gamutmap_get_max(data, tmp);
      if(step < 1e-10f) goto fail;
      if(max > 0.0 && !gamutmap_get_normal(data, tmp, n))
      {
        for(int k=0;k<3;k++) newlab[k] = lab[k];
        break;
      }
      ii++;
      step *= .5f; // keep reducing until we get a max/normal inside spectral locus
    }
    // we're losing 1% of converged samples here, but for rare cases it also helps..
    // if(ii==0) step = fminf(1.0f, step * 2.0f);

    for(int k=0;k<3;k++) xyz[k] = tmp[k] / max;
    dt_XYZ_to_Lab(xyz, lab);
    float d2[3] = {lab[0] - in_lab[0], lab[1] - in_lab[1], lab[2] - in_lab[2]};
    last_deltaE2 = dot_dd;
    dot_dd = dotproduct(d2, d2);
    dot_dn = dotproduct(d2, n);
  }
  assert(xyz[0] == xyz[0]);
  assert(xyz[1] == xyz[1]);
  assert(xyz[2] == xyz[2]);
  for(int k=0;k<3;k++) out_xyz[k] = xyz[k];
  return 0;
fail:
  fprintf(stderr, "diverged %g, %g, %g\n", in_xyz[0], in_xyz[1], in_xyz[2]);
  for(int k=0;k<3;k++) out_xyz[k] = 0.0f;
  return 1;
}

