#pragma once

#include <math.h>
#include <stdint.h>
#include <assert.h>

// stolen from darktable, GPL

static inline float
cbrt_5f(float f)
{
  uint32_t* p = (uint32_t*)&f;
  *p = *p/3 + 709921077;
  return f;
}

static inline float
cbrta_halleyf(const float a, const float R)
{
  const float a3 = a*a*a;
  const float b = a * (a3 + R + R) / (a3 + a3 + R);
  return b;
}

static inline float
lab_f(const float x)
{
  const float epsilon = 216.0f/24389.0f;
  const float kappa   = 24389.0f/27.0f;
  if(x > epsilon)
  {
    // approximate cbrtf(x):
    const float a = cbrt_5f(x);
    return cbrta_halleyf(a, x);
  }
  else return (kappa*x + 16.0f)/116.0f;
}

static inline void
dt_XYZ_to_Lab(const float *XYZ, float *Lab)
{
  const float d50[3] = { 0.9642f, 1.0f, 0.8249f };
  const float f[3] = { lab_f(XYZ[0]/d50[0]), lab_f(XYZ[1]/d50[1]), lab_f(XYZ[2]/d50[2]) };
  Lab[0] = 116.0f * f[1] - 16.0f;
  Lab[1] = 500.0f*(f[0] - f[1]);
  Lab[2] = 200.0f*(f[1] - f[2]);
}

static inline float
lab_f_inv(const float x)
{
  const float epsilon = 0.20689655172413796f; // cbrtf(216.0f/24389.0f);
  const float kappa   = 24389.0f/27.0f;
  if(x > epsilon) return x*x*x;
  else return (116.0f*x - 16.0f)/kappa;
}

static inline void
dt_Lab_to_XYZ(const float *Lab, float *XYZ)
{
  const float d50[3] = { 0.9642f, 1.0f, 0.8249f };
  const float fy = (Lab[0] + 16.0f)/116.0f;
  const float fx = Lab[1]/500.0f + fy;
  const float fz = fy - Lab[2]/200.0f;
  XYZ[0] = d50[0]*lab_f_inv(fx);
  XYZ[1] = d50[1]*lab_f_inv(fy);
  XYZ[2] = d50[2]*lab_f_inv(fz);
}

void deltaE_gradient(float diff, float *vis)
{
  assert(diff >= 0.0f);
  const float JND = 2.3f;
  const float col[][3] = {
    {0.0f, 0.0f, 0.0f},
    {0.0f, 1.0f, 0.0f},
    {1.0f, 1.0f, 0.0f},
    {1.0f, 0.0f, 0.0f}};
  const int num_col = sizeof(col)/sizeof(col[0]);
  for(int c=0;c<num_col-1;c++)
  {
    if(c == num_col-2 || diff <= (c+1)*JND)
    {
      const float t = fminf(1.0f, (diff - c*JND)/JND);
      for(int k=0;k<3;k++)
        vis[k] = t * col[c+1][k] + (1.0f-t) * col[c][k];
      return;
    }
  }
}

void diffvis(const float *xyz0, const float *xyz1, float *vis)
{
  float lab0[3], lab1[3];
  dt_XYZ_to_Lab(xyz0, lab0);
  dt_XYZ_to_Lab(xyz1, lab1);

  // delta E cie 76 - 2.3 is just noticable difference
  float diff = sqrtf(
      (lab0[0]-lab1[0]) * (lab0[0]-lab1[0]) +
      (lab0[1]-lab1[1]) * (lab0[1]-lab1[1]) +
      (lab0[2]-lab1[2]) * (lab0[2]-lab1[2]));
  deltaE_gradient(diff, vis);
}
